from typing import TypeVar
from .exceptions import *

ListOrDict = TypeVar('ListOrDict', list, dict)
ListOrString = TypeVar('ListOrString', list, str)


class Stack:
    """
        The Common Stack Data Structure That You Can Use Here.
    """

    def __init__(self, inf=1, capacity=0):
        """
            Arguements :
                * inf=1, # Set If Stack Is Infinite Or Not. Default Is Infinite Stack.
                * capacity=0 # Set Capacity In Case Of limited Stack.
        """
        self.inf = inf
        if inf:
            self.__storage = []
        else:
            self.ind = 0
            self.__storage = [0 for _ in range(capacity)]
            if capacity == 0:
                raise ValueError(
                    "If You Set Inf Mode To Limited You Must Set Capacity For Stack.")
            self.capacity = capacity

    @property
    def storage(self):
        return self.__storage

    def push(self, elem):
        """
            Push The Value To Stack
        """
        if self.inf:
            self.__storage.append(elem)
        else:
            if self.ind == self.capacity:
                raise StackError("Stack Is Full !")
            self.__storage[self.ind] = elem
            self.ind += 1

    def pop(self):
        """
            Pop The Value From Stack ( Returns The Removed Value )
        """
        if self.__storage:
            return self.__storage.pop()
        else:
            raise StackError("Stack Is Empty !")

    def display(self):
        """
            Display Stack Data
        """
        return self.storage


class LIFO(Stack):
    """
        You Can Use This Insted Of Stack. All Methods Of Stack Are Available Here.
    """
    pass


class Queue:
    """
        The Common Queue Data Structure That You Can Use Here.
    """

    def __init__(self, inf=1, capacity=0):
        self.inf = inf
        if inf:
            self.__storage = []
        else:
            self.ind = 0
            self.__storage = [0 for _ in range(capacity)]
            if capacity == 0:
                raise ValueError(
                    "If You Set Inf Mode To Limited You Must Set Capacity For Queue.")
            self.capacity = capacity

    @property
    def storage(self):
        return self.__storage

    def add(self, elem):
        """
            Add Value To Queue.
        """
        if self.inf:
            self.__storage.append(elem)
        else:
            if self.ind == self.capacity:
                raise QueueError("Queue Is Full !")
            self.__storage[self.ind] = elem
            self.ind += 1

    def remove(self):
        """
            Remove From Queue ( Returns The Removed Value )
        """
        if self.__storage:
            return self.__storage.pop(0)
        else:
            raise QueueError("Queue Is Empty !")

    def display(self):
        """
            Display The Queue.
        """
        return self.storage


class FIFO(Queue):
    """
        You Can Use This Insted Of Queue. All Methods Of Queue Are Available Here.
    """
    pass


class PyshaList(list):
    """
        PyshaList That Have Some Advanteges To Normal List.
    """

    def __str__(self):
        s = "["
        for i in self:
            s += " " + str(i) + " ,"
        s = s[:-1] + "]"
        return s

    def __repr__(self):
        s = "["
        for i in self:
            s += " " + str(i) + " ,"
        s = s[:-1] + "]"
        return s

    def lshift(self, times):
        """
            Left Shift In Times That You Specify
        """
        for _ in range(times):
            self = self[1:] + [self[0]]

    def rshift(self, times):
        """
            Right Shift In Times That You Specify
        """
        for _ in range(times):
            self = [self[-1]] + self[:-1]

    # Just String,Int And Boolean Are Handled Here.
    def count_deep(self, tar):
        """
Deep Count. If You Pass A list Of String, It Will Check If The tar ( Argument That Passed To This Function )
Is Inside The list itself or is inside the elements of list or not. If you pass a list with complex dimentions
it will count inside of all of elements even insidest element ! and it will count it for you.
        """
        if not (type(tar) == int or type(tar) == str or type(tar) == bool):
            raise TypeError(
                "CountDeep Just Supports String,Integer And Boolean Types.")
        if not any(list(map(lambda x: type(x) == list or type(x) == set or type(x) == tuple, self))):
            if type(tar) == str and any(list(map(lambda x: type(x) == str, self))):
                counted = 0
                for index in range(len(self)):
                    try:
                        counted += self[index].count(tar)
                    except:
                        continue
                return counted
            return self.count(tar)
        indexes = Stack()
        lists = Stack()
        target_list = self
        current_index = 0
        counted = 0
        while current_index != len(self):
            has_new_list = False
            for i in target_list:
                if type(tar) == type(i):
                    if tar == i:
                        counted += 1
                elif type(tar) != list:
                    indexes.push(current_index)
                    lists.push(target_list)
                    target_list = i.copy()
                    current_index = 0
                    has_new_list = True
                    break
                current_index += 1
            if not has_new_list:
                current_index = indexes.pop() + 1
                target_list = lists.pop()
        return counted

    def shift_index(self, index, left_shift=True, how_many=1):
        """
You Can Shift Your Array Better That Just lshift And rshift Methods. You Can Define Which Index To
Shift, Type Of Shift ( left or right ) And how_many To Shift.

Arguments :
* index, # Index That You Want To Shift
* left_shift=True, # Left Or Right Shift That You Can Specifiy. Default Is Left Shift
* how_many=1 # How Many Shifts That You Want. Default Is One
        """
        for _ in range(how_many):
            if not left_shift:
                if index == len(self) - 1:
                    self.rshift(1)
                else:
                    self[index], self[index +
                                      1] = self[index + 1], self[index]
                index = (index + 1) % len(self)
            else:
                if index == 0:
                    self.lshift(1)
                else:
                    self[index], self[index -
                                      1] = self[index - 1], self[index]
                index = (index - 1) % len(self)

    def __setitem__(self, k: int, v: object) -> None:
        if type(k) == int:
            if k > len(self):
                self.extend([None for _ in range(k - len(self) + 1)])
                super(PyshaList, self).__setitem__(k, v)
            else:
                super(PyshaList, self).__setitem__(k, v)
        elif type(k) == tuple:
            self.set_array(k, v)
        else:
            raise TypeError("Type Of Index Must Be int.")

    # List Operators + Some Beautiful Things :) #

    def __add__(self, o: object) -> list:
        if type(o) == list:
            cp = self.copy()
            cp.extend(o)
            return cp
        else:
            tar = []
            for i in self:
                try:
                    tar.append(i + o)
                except:
                    tar.append(str(i) + " " + str(o))
            return tar

    def __iadd__(self, o: object) -> list:
        if type(o) == list:
            self.extend(o)
        else:
            for i in self:
                try:
                    self.append(i + o)
                except:
                    self.append(str(i) + " " + str(o))

    def __sub__(self, o: object) -> list:
        if type(o) == list:
            tar = []
            for i in self:
                if not i in o:
                    tar.append(i)
            return tar
        if type(o) == int or type(o) == float:
            tar = []
            for i in self:
                try:
                    tar.append(i - o)
                except:
                    pass
            return tar

    # **
    def __pow__(self, o: object) -> list:
        return list(map(lambda elem: elem ** o, self))

    # /
    def __truediv__(self, o: object) -> list:
        return list(map(lambda elem: elem/o, self))

    # //
    def __floordiv__(self, o: object) -> list:
        return list(map(lambda elem: elem//o, self))

    # <<
    def __lshift__(self, o: object) -> list:
        if type(o) == int:
            return self.lshift(o)
        elif type(o) == dict:
            for i, j in o.items():
                if type(i) == int:
                    self.shift_index(i, True, j)
                elif type(i) == tuple or type(i) == list or type(i) == set:
                    for a in i:
                        self.shift_index(a, True, j)
                else:
                    raise TypeError("Type Must Be int,tuple,list or set.")
        else:
            raise TypeError("Type Must Be int Or dict.")

    # >>
    def __rshift__(self, o: object) -> list:
        if type(o) == int:
            return self.rshift(o)
        elif type(o) == dict:
            for i, j in o.items():
                if type(i) == int:
                    self.shift_index(i, False, j)
                elif type(i) == tuple or type(i) == list or type(i) == set:
                    for a in i:
                        self.shift_index(a, False, j)
                else:
                    raise TypeError("Type Must Be int,tuple,list or set.")
        else:
            raise TypeError("Type Must Be int Or dict.")

    # %
    def __mod__(self, o: object) -> list:
        return list(map(lambda elem: elem % o, self))

    @staticmethod
    def make_array(dimentions):
        """
Make Array In Complex Dimentions. You Can Use This Method Like : 
```PyshaList.make_array([1,2,4,5,4,32])```
        """
        ans = [None for _ in range(dimentions[-1])]
        for i in dimentions[:-1][::-1]:
            ans = [ans for _ in range(i)]
        return ans

    def set_array(self, dimentions, value):
        """
Set Array In Complex Dimentions. You Can Use This Method Like : 
```
array = PyshaList.make_array([1,2,4,5,4,32]) # Just A Complex Array
array.set_array([0,2,4,0,4,3],"test")
```
Or Like This :
```array[0,2,4,0,4,3] = "test"```
Cool, Right ? :)
        """
        array = self
        for i in dimentions:
            array = array[i]
        array = value
        return array


class PyshaString(str):
    """
        In PyshaString You Have Lots Of Methods To Do Some Things.
    """

    def inside(self, data_to_search: ListOrDict, search_mode="key", return_type=bool):
        """
In This Function You Can Search That If A String Exists In Dict ( Or Inside Of All Elements ).
Or If A String Exists In List And All Of Its Elements.

Args :
* data_to_search ->  Data That Wanted To Search In That. For Example
                    Search In List : ["test", "some", "Hello World"] Or
                    Search In dict : {"test": "Hello World", "some": "Hello World"} 
                    -> ( search_mode must be "value" or "both" in dict input )
                    if string is Hello It Will Return True !

* search_mode ->  If You Passed A Dictionary You Can Choose Between "key" Or "value"
                Or "all"/"both" To Search In Both Of Them. It Does Not Do Anything If
                You Passed A List.

* return_type ->  If You Passed A Dictionary You Can Choose Between bool and dict or ...
                if you pass bool, it will return True or False If It Finds Your String.
                if you pass dict, it will return a dict with key and value of found element.
                if you pass pass anything else, it will return oposite of search_mode. For
                Example If You passed "key" And You Found Your String, It Will Return "value".

        """
        if type(data_to_search) == list:
            if self in data_to_search:
                return True
            for i in data_to_search:
                if self in i:
                    return True
            return False
        elif type(data_to_search) == dict:
            ans = {}
            if return_type == dict:
                ans = {}
            else:
                ans = []
            for i, j in data_to_search.items():
                if search_mode == "key":
                    if self in i:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
                elif search_mode == "value":
                    if self in j:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
                else:
                    if self in j or self in i:
                        if return_type == bool:
                            return True
                        elif return_type == dict:
                            ans[i] = j
                        else:
                            ans.append(j)
            if return_type != bool:
                return ans
            return False
        else:
            raise TypeError("Type Must Be list Or dict.")

    def find_first(self, data_to_search: list, return_type=int):
        """
In This Function You Can Search That an String Contains Some Other String From A List.

Args :
* data_to_search ->  Strings That You Want To Search For. For Example
                    Search In List : [" ", ":", ","] And If Your String Is 
                    "Hello, Im Arshia" It will Return 5 Thats Index Of ",".
                    If Your String Is "Arshia:Hey" It Will Return 6 Thats Index Of ":".

* return_type ->  You Can pass int To Return Index Of Found String. Or Pass Other
                Things Like str To Return Found String.

        """
        for c, i in enumerate(self):
            if i in data_to_search:
                if return_type == int:
                    return c
                else:
                    return i

    def find_last(self, data_to_search: list, return_type=int):
        """
This Function Is Exactly find_first But It Will Return Last Index Of Found String.

Args :
* data_to_search ->  Strings That You Want To Search For. For Example
                    Search In List : [" ", ":", ","] And If Your String Is 
                    "Hello, Im Arshia" It will Return 10 Thats Index Of " ".
                    If Your String Is "Arshia:Hey" It Will Return 6 Thats Index Of ":".

* return_type ->  You Can pass int To Return Index Of Found String. Or Pass Other
                Things Like str To Return Found String.

        """
        for c, i in enumerate(self[::-1]):
            if i in data_to_search:
                if return_type == int:
                    return len(self) - c - 1
                else:
                    return i

    def replace_array(self, data_to_replace: list, replace_with: ListOrString):
        """
In This Function You Can Replace An Array With Another Array ! ( in a string ).
Its Possible That You Replace An Array Of String With A Single String Too.

Args :
* data_to_replace ->  Data That You Wanted To Be Replaced For Example If Data Is 
                    ["Hello", "World"] And replace_with Argument Is ["Hi", "There"]
                    All Of "Hello" And "World" Will Be Replaced With "Hi" And "There".

* replace_with ->  Described In data_to_replace Argument.

        """
        if type(replace_with) == str:
            for i in data_to_replace:
                self = self.replace(i, replace_with)
        else:
            for c, i in enumerate(data_to_replace):
                self = self.replace(i, replace_with[c])
        return self

    def replace_dict(self, data_to_replace: dict):
        """
This Is replace_array But You Can Pass Dictionary To Replace With.

Args :
* data_to_replace ->  A Dictionary Of Data To Replace. For Example 
                    {"Hello": "Hi", "World": "There"} Will Replace "Hello" With "Hi"
                    And "World" With "There". ( All Of String )

        """
        for j, i in data_to_replace.items():
            self = self.replace(j, i)
        return self

    def exists(self, data_to_search: list, line_start=-1, line_end=-1, start_index=-1, end_index=-1):
        """
This Function Will Check If A List Of Strings Exists In Another String. In Specific
Part Of It Or Not. It Will Get a Part Of String And Will Check It With A List Of Strings.

Args :
* data_to_replace ->  A List Of Data That You Want To Check If Exists In Another String.

* line_start ->  Line Start Number ( Starts With 0 )

* line_end ->  Line End Number ( Starts With 0 )

* start_index ->  Start Index Number ( Starts With 0 )

* end_index ->  End Index Number ( Starts With 0 )

        """
        if line_start == -1:
            line_start = 0
        if line_end == -1:
            line_end = len(self)
        if start_index == -1:
            start_index = 0
        if end_index == -1:
            end_index = len(self)
        data_to_search = '\n'.join(
            data_to_search.split('\n')[line_start:line_end])[start_index:end_index]
        for i in data_to_search:
            if i in self:
                return True
        return False

    def replace_with_escape(self, string_to_find, to_replace, escape_char='\\', remove_escapes=False):
        """
This Function Will Replace Any String That You Want But You Can Define Some Parts
Not To Replace With Define Escapes !

Args :
* string_to_find ->  String That You Want To Search For In String.

* to_replace ->  String That You Want To Be Replaced With string_to_find.

* escape_char ->  Escape Character That You Want To Define. ( Default Is \ Character )

* remove_escapes ->  You Can Define That After Replacing, Escape Characters Should Remove
                    Or Not. ( Default Is False )
        """
        last_index = self.find(string_to_find, 0)
        while last_index != -1:
            if self[last_index - 1] == escape_char:
                if remove_escapes:
                    self = self[:last_index-1] + self[last_index:]
                last_index = self.find(string_to_find, last_index)
                continue
            self = self[:last_index] + to_replace + \
                self[last_index + len(string_to_find):]
            last_index = self.find(string_to_find, last_index + 1)
        return self

    def __ilshift__(self, other):
        """
            This Function Is "<<=" Operator. It Will Append Another String To The End Of This String.
        """
        self = self + other
        return self

    def __iadd__(self, other):
        """
            This Function Is "+=" Operator. It Will Append Another String To The End Of This String.
        """
        self = self + other
        return self

    def __irshift__(self, other):
        """
            This Function Is ">>=" Operator. It Will Append Another String To Start Of This String.
        """
        self = other + self
        return self

    def __lshift__(self, other):
        """
            This Function Is "<<" Operator. It Will Append Another String To The End Of This String. ( Not Set )
        """
        return self + other

    def __rshift__(self, other):
        """
            This Function Is ">>" Operator. It Will Append Another String To Start Of This String. ( Not Set )
        """
        return other + self

    def __isub__(self, other):
        """
            This Function Is "-=" Operator. It Will Remove A String From This String.
        """
        self = self.replace(other, "")
        return self

    def __sub__(self, other):
        """
            This Function Is "-" Operator. It Will Remove A String From This String. ( Not Set )
        """
        return self.replace(other, "")


class PyshaDict(dict):
    """
Thats Just Normal Dictionary Just With 2 Or 3 More Option.
PyshaDict Operators :
* ~myDict # Returns Dictionary Inverse
* myDict - "elem" # Returns Dicionary With Removed "elem" Key.
    """
    @property
    def inverse(self):
        """
You Can Get inverse Of A Dictionary. For Example Dict -> `{"nickname":"MrMegaExpert"}`, Inverse -> `{"MrMegaExpert":"nickname"}`
You Can Set And Change Dicitonary In Inverse Mode ! You Can Get Inverse Of A Dictionary In 2 Ways :
* myDict.inverse
* ~myDict
        """
        return PyshaDict({v: k for k, v in self.items()})

    @inverse.setter
    def inverse(self, value):
        value = {v: k for k, v in value.items()}
        self.update(value)

    def __invert__(self):
        return self.inverse

    def __isub__(self, other):
        del self[other]
        return self

    def __sub__(self, other):
        ans = self.copy()
        del ans[other]
        return ans



from enum import Enum
from inspect import signature
import platform
from .data_structures import PyshaString
from .functions import *
from os import environ, popen, remove


class command:
    """
        This Class Will Be Use For Fun Part Of This Framework :)
    """

    def __init__(self):
        self.arr_data = []
        self.dic_tar = {}
        self._mode = 'p'

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, m):
        if m not in ['p', 'i']:
            raise ModeError(
                "Modes Should Be 'p' Or 'i'. Default Is 'p' / Print Mode.")
        else:
            self._mode = m

    def loop(self, cmd, mode='p', c=True):
        """
2 Types Of Loop Currently Exists. i(Input) and p(Print) Mode.
In print mode you can make a print loop like that :
```<i:3>{You Are In _i_ Loop}```
And This Will Be Printed Like That : 
```You Are In 0 Loop```
```You Are In 1 Loop```
```You Are In 2 Loop```

And In Input Mode You Can Do This :
```<i:3>{Enter _i_ Number : }```
It Will Get 3 Input Like This :
```Enter 0 Number : (input0)```
```Enter 1 Number : (input1)```
```Enter 2 Number : (input2)```
And Will Return `[input0,input1,input2]`

More Options :
```<i:3>{Enter _i++_ Number : }```
```<i:3>{Enter _i--_ Number : }```
```<i:2,j:2>{Enter Row _i++_ And Column _j++_  Number : }```
```<i:2>{Enter (Fore.CYAN)[_i++_] Number : }```

Notes :
* Just Maded Maximum For 2 Variables Not More.
        """
        if self.mode != 'p':
            mode = self.mode
        if mode == "p":  # Print Mode
            self.dic_tar.clear()
            self.arr_data.clear()
            alls = cmd.split(">")[0][1:]
            string = cmd.split("{")[1][:-1]
            for i in alls.split(','):
                itemp = i.split(':')
                self.arr_data.append(itemp[0].strip())
                self.dic_tar[itemp[0].strip()] = int(itemp[1].strip())
            if len(self.dic_tar.keys()) >= 4:
                raise LoopError(
                    "Loops With 4 Or More Variables Are Not Supported In This Version.")
            tars = make_possibles(self.dic_tar)
            for i in tars:
                temp = {}
                if not (type(i) == int) and len(i) == 2:
                    temp = {self.arr_data[0]: i[0], self.arr_data[1]: i[1]}
                elif not (type(i) == int) and len(self.dic_tar) == 3:
                    temp = {
                        self.arr_data[0]: i[0], self.arr_data[1]: i[1], self.arr_data[2]: i[2]}
                else:
                    temp = {self.arr_data[0]: i}
                display(temp, string, c)
        elif mode == 'i':  # input Mode
            self.dic_tar.clear()
            self.arr_data.clear()
            alls = cmd.split(">")[0][1:]
            string = cmd.split("{")[1][:-1]
            for i in alls.split(','):
                itemp = i.split(':')
                self.arr_data.append(itemp[0].strip())
                self.dic_tar[itemp[0].strip()] = int(itemp[1].strip())
            if len(self.dic_tar.keys()) >= 4:
                raise LoopError(
                    "Loops With 4 Or More Variables Are Not Supported In This Version.")
            tars = make_possibles(self.dic_tar)
            x = 0
            y = 0
            z = 0
            ls = []
            if len(self.dic_tar) == 2:
                x = (list(self.dic_tar.values())[0])
                y = (list(self.dic_tar.values())[1])
                ls = [["" for _ in range(y)] for __ in range(x)]
            elif len(self.dic_tar) == 1:
                x = (list(self.dic_tar.values())[0])
                ls = ["" for _ in range(x)]
            elif len(self.dic_tar) == 3:
                x = (list(self.dic_tar.values())[0])
                y = (list(self.dic_tar.values())[1])
                z = (list(self.dic_tar.values())[2])
                ls = [[["" for _ in range(z)] for __ in range(y)]
                      for ___ in range(x)]
            x = 0
            y = 0
            z = 0
            for i in tars:
                temp = {}
                if not (type(i) == int) and len(i) == 2:
                    temp = {self.arr_data[0]: i[0], self.arr_data[1]: i[1]}
                    try:
                        ls[x][y] = get_ans(temp, string, c)
                        ls[x][y + 1]
                        y += 1
                    except:
                        try:
                            ls[x + 1][y]
                            x += 1
                            y = 0
                        except:
                            pass
                elif not (type(i) == int) and len(i) == 3:
                    temp = {
                        self.arr_data[0]: i[0], self.arr_data[1]: i[1], self.arr_data[2]: i[2]}
                    try:
                        ls[x][y][z] = get_ans(temp, string, c)
                        ls[x][y][z + 1]
                        z += 1
                    except:
                        try:
                            ls[x][y + 1][z]
                            y += 1
                            z = 0
                        except:
                            try:
                                ls[x + 1][y][z]
                                x += 1
                                y = 0
                                z = 0
                            except:
                                pass
                else:
                    temp = {self.arr_data[0]: i}
                    try:
                        ls[x] = get_ans(temp, string, c)
                        ls[x]
                        x += 1
                    except:
                        pass

            return ls
        else:
            raise ValueError("Mode Should Be 'p' Or 'i' !")

    def condition(self, check, p=True, **kwargs):
        """
This Function Act As c++ One Line Condition.
You Can Use This Function Like :

`condition('i > j ? "i Is Grater" : "j Is Grater"',i=20,j=23)`

Output Will Be :

`"j Is Grater"`

Notes & Warnings :

`-> '(Quotes) Are Not Working, Always Use "(Double Quotes)`
        """
        if not re.match("""(\"?)[ -~]+(\"?)(>|<|==|>=|<=)(\"?)[ -~]+(\"?)\?(\"?)[ -~]+(\"?):\W{,1}(\"?)[!-~]+(\"?)""", ''.join(check.split(' '))):
            raise ConditionError("An Error In Parsing Condition :(")
        if "eval" in check:
            if not check[check.find("eval") - 1:].startswith("\eval"):
                raise SecurityError("Using This Funcion Is Forbidden.")
        if "exec" in check:
            if not check[check.find("exec") - 1:].startswith("\exec"):
                raise SecurityError("Using This Funcion Is Forbidden.")
        index_of_ok_sign = not_between_index("?", check)
        p1 = index_split(check, index_of_ok_sign)[0].strip()
        p2 = index_split(check, index_of_ok_sign)[1].strip()
        ans = eval(p1, kwargs, {})
        index_of_ok_sign = not_between_index(":", p2)
        ok = index_split(p2, index_of_ok_sign)[0].strip()
        not_ok = index_split(p2, index_of_ok_sign)[1].strip()
        if ans:
            if p:
                print(eval(ok, kwargs, {}))
            else:
                return eval(ok, kwargs, {})
        else:
            if p:
                print(eval(not_ok, kwargs, {}))
            else:
                return eval(not_ok, kwargs, {})

    def exe(self, string, priority=0, strip_response=True):
        """
In Progress But This Function Should Get Text Between () Of $(cmd)
And Run It In Terminal Then Replace The Output With $(cmd) And Should
Get #(cmd) Too And Run It As Python Code And Replace This Output Too.
    Problem Now :
        Use This 2 Thing Together.

Args : 

* string          ->    The String Pattern That You Want To Make.

* priority        ->    If Its 0 First System Commands Will Replace Then
                        Python Commands Will Replace. And If Its 1 First
                        Python Commands Will Replace Then System Commands
                        Will Be Replaced. Default Is 0.

* strip_response  ->     Default Is True. It Wanted To Know That Your
                            String That Replaced Should Be striped ? ( Remove
                            Addtion New Lines And Spaces In Start And End Of 
                            String ? )

Example :

* exe("You Current Directory :\n $(dir)\nFinished :)")
        """
        s = re.compile(
            "\$\([A-Za-z0-9_ \$\&\!\@\#\%\^\*\+\=\-\]\[\~\?\'\"\;\:]+\)")
        p = re.compile(
            "\#\([A-Za-z0-9_ \$\&\!\@\#\%\^\*\+\=\-\]\[\~\?\'\"\;\:]+\)")
        if priority == 0:
            # First System Then Python
            for cmd in s.finditer(string):
                cmds = cmd.group()[2:-1]
                ans = popen(cmds).read()
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            for cmd in p.finditer(string):
                cmds = cmd.group()[2:-1]
                f = open("temp.py", "w")
                f.write(cmds)
                f.close()
                ans = popen(environ.get("python") + " temp.py").read()
                remove("temp.py")
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            if strip_response:
                return string.strip()
            else:
                return string
        elif priority == 1:
            # First Python Then System
            for cmd in p.finditer(string):
                cmds = cmd.group()[2:-1]
                f = open("temp.py", "w")
                f.write(cmds)
                f.close()
                ans = popen(environ.get("python") + " temp.py").read()
                remove("temp.py")
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            for cmd in s.finditer(string):
                cmds = cmd.group()[2:-1]
                ans = popen(cmds).read()
                string = string[:cmd.start()] + ans + \
                    string[cmd.start() + len(cmd.group()):]
            if strip_response:
                return string.strip()
            else:
                return string
        else:
            ValueError(
                "Priority Should Be 0 (System First) Or 1 (Python First).")


class Switch:
    """
        The Famous Switch-Case :)
    """

    def __init__(self, var):
        self.var = var

    def case(self, tar, func, args=tuple()):
        if self.var == tar or (hasattr(tar, '__name__') and tar.__name__ == "Case" and tar == self.var) or (tar == self.var) or (hasattr(tar, '__name__') and tar.__name__ == "Default"):
            func(*args)

    def __getitem__(self, all_funcs):
        dic = {}
        for i in all_funcs:
            dic[i.start] = i.stop
        for i, j in dic.items():
            if self.var == i or (hasattr(i, '__name__') and i.__name__ == "Case" and i.tar == self.var) or (hasattr(i, 'tar') and i.tar == self.var) or (hasattr(i, '__name__') and i.__name__ == "Default"):
                if hasattr(j, '__call__'):
                    j()
                    return
                elif type(j) == str:
                    parts = j.split('\n')
                    if parts[0].strip():
                        t = parts[0].index(j.split('\n')[0].strip())
                    else:
                        t = parts[1].index(j.split('\n')[1].strip())
                    ok_code = ""
                    for i in j.split('\n'):
                        ok_code += i[t:] + "\n"
                    b = compile(ok_code, 'user_code', 'exec')
                    a = exec(b)
                    if a != None:
                        print(a)
                    return
                if len(j) == 3:
                    j[0](*j[1], **j[2])
                    return
                elif len(j) == 2:
                    j[0](*j[1])
                    return
                elif len(j) == 1:
                    j[0]()
                    return

    def cases(self, all_funcs):
        """
This Function Should Get A Dictionary That Key Is Case And Value Is Function That Will Run On That
Case. You Can Pass Lots Of Types For Example :
```
    Case(5):
        [
            func,
            arguments,
            kwargs
        ]
```
Or :
```
    Case(1):
        " \
            print("yay")
        "
```
Or :
```
    Case(6):
        lambda x,y:
            print(x,y)
```
For More Models You Can See Examples.
        """
        for i, j in all_funcs.items():
            if self.var == i or (hasattr(i, '__name__') and i.__name__ == "Case" and i.tar == self.var) or (hasattr(i, 'tar') and i.tar == self.var) or (hasattr(i, '__name__') and i.__name__ == "Default"):
                if hasattr(j, '__call__'):
                    j()
                    return
                elif type(j) == str:
                    parts = j.split('\n')
                    if parts[0].strip():
                        t = parts[0].index(j.split('\n')[0].strip())
                    else:
                        t = parts[1].index(j.split('\n')[1].strip())
                    ok_code = ""
                    for i in j.split('\n'):
                        ok_code += i[t:] + "\n"
                    b = compile(ok_code, 'user_code', 'exec')
                    a = exec(b)
                    if a != None:
                        print(a)
                    return
                if len(j) == 3:
                    j[0](*j[1], **j[2])
                    return
                elif len(j) == 2:
                    j[0](*j[1])
                    return
                elif len(j) == 1:
                    j[0]()
                    return


class Default:
    """
        Default Case Of Switch-Case. Its Not Doing Anything But You Can Use It On Your Switch Cases.
    """
    pass


class Case:
    """
        Define Case Of Switch-Case. Just In Case You Want To Use It Like Case(x).
    """

    def __init__(self, tar):
        self.tar = tar


class Vars:
    """
        In Case You Want To Define Some Variables In lambdas, You Can Set Attributes Of This Class.
    """
    pass


class PercentPrinter:
    """
        Pretty Percent Printer For Loading Some Progress Or Downloading :)
        In Fact, Its Progressbar Itself :)
    """

    def __init__(self, chars=100, pass_color=fore["reset"], loading_color=fore["reset"]):
        """
Arguments :
* chars=50, # Characters Of Progressbar
* pass_color=fore["reset"], # The Color Of Passed Part
* loading_color=fore["reset"], # The Color Of Loading Part ( Still Not Fulled Part )
        """
        self._percent = 0
        self.chars = chars
        self.pass_color = pass_color
        self.loading_color = loading_color
        self.char_ok = '#'
        self.char_loading = '-'

    def config(self, char_ok, char_loading):
        """
            Set char_ok And char_loading For Progressbar.
        """
        self.char_ok = char_ok
        self.char_loading = char_loading

    def show(self, char_ok='#', char_loading='-', end='\n'):
        """
            This Function Just Shows The Progressbar.
        """
        if char_ok == '#':
            if self.char_ok != '#':
                char_ok = self.char_ok
        if char_loading == '-':
            if self.char_loading != '-':
                char_loading = self.char_loading

        num = self._percent
        string = "\r"
        string += (self.pass_color + char_ok + Fore.RESET) * \
            int(num * (self.chars / 100))
        string += (self.loading_color + char_loading + Fore.RESET) * \
            int((100 - num) * (self.chars / 100))
        print(string, end=end)

    @property
    def percent(self):
        return self._percent

    @percent.setter
    def percent(self, p):
        if p > 100:
            raise ValueError("Value Most Be Less Than 100 !")
        elif p < 0:
            raise ValueError("Value Most Be More Than 0 !")
        else:
            self._percent = p

    def increase(self, p=1, show=True):
        """
            Increses Progressbar In Percents That You Set.
        """
        self._percent += p
        if show:
            self.show()

    def finish(self, show=True):
        """
            Finishes The Progressbar.
        """
        self._percent = 100
        if show:
            self.show()


class CrossPlatformer:
    """
        This Class Will Make An Environment For You To Have Easier And Better Experience :)
    """

    def __init__(self, supported_os=["linux", "windows", "mac"]) -> None:
        """
Arguments : 
* supported_os # You Can Set Supported Oses. Default is ["linux", "windows", "mac"]
        """
        self.user_os = platform.system().strip().lower()
        self.all_commands = {}
        self.supported_os = supported_os
        self.platforms = {
            "windows": ["windows", "win32"],
            "linux": ["linux", "linux2"],
            "mac": ["mac", "darwin"]
        }

    def get_os_name(self):
        """Get User Os Name"""
        if self.user_os == "linux" or self.user_os == "linux2":
            return "Linux"
        elif self.user_os == "darwin":
            return "Mac"
        elif self.user_os == "win32" or self.user_os == "windows":
            return "Windows"
        else:
            return "Unknown"

    def is_os_supported(self, os_name):
        """Check If An Os Is Supported"""
        for i, j in self.platforms.items():
            if os_name.lower().strip() in j or os_name.lower().strip() == i:
                return True
        return False

    def get_os_version(self):
        """Get User Os Version"""
        if self.user_os == "linux" or self.user_os == "linux2":
            return platform.linux_distribution()[0]
        elif self.user_os == "darwin":
            return platform.mac_ver()[0]
        elif self.user_os == "win32" or self.user_os == "windows":
            return platform.win32_ver()[0]
        else:
            return "Unknown"

    def get_os_command(self, name):
        """
Get Command For User Os. Arguments :
* name : Command Name
For Example :

If You Set add_os_commands("clear", {"linux": "clear", "windows": "cls", "mac": "clear"}) :
* get_os_command("clear")

Will Return You The Command Depends On Your Os.
        """
        try:
            return self.all_commands[name][self.user_os]
        except:
            return None

    def add_os_commands(self, name, commands: dict):
        """
Set Command For User Os. Arguments Are : 
* name : Name Of Command
* commands : Dictionary Of Commands
For Example : 
* add_os_commands("git", {"linux": "git", "windows": "git.exe"})
Or : 
* add_os_commands("clear", {"linux": "clear", "windows": "cls", "mac": "clear"})
        """
        for i in commands:
            if not self.is_os_supported(i):
                raise ValueError(f"This OS ({i}) Is Not Supported !")
        self.all_commands[name] = commands

    def __getitem__(self, name):
        return self.get_os_command(name)

    def __setitem__(self, name, commands):
        self.add_os_commands(name, commands)


class interface:
    """
You Can Make Interfaces ( Inspired From PHP ! ) .
In Fact You Can Define An Structure That All Of Classes That Extends That Interface Must Have This
Structure That You Defined.

You Can Use It Like Decorator. For Example : 

```
@interface
class UserStructure:
    name = None
    family = None
    username = None
    password = None

@interface(UserStructure)
class SpecialUsers:
    name = None
    family = None
    username = None
    password = None

@interface(UserStructure)
class Users:
    name = None
    family = None
    username = None
    password = None
```
    """
    instances: dict = {}

    def __init__(self, target_class):
        self.is_parent_interface = False
        try:
            class_name = target_class.__name__
        except:
            try:
                class_name = target_class._class.__name__
            except:
                raise ValueError("You Should Pass className.")
        if (type(target_class) == type):
            self._class = target_class
            self.is_parent_interface = True
            interface.instances[class_name] = self
        else:
            self._parent = target_class
            self.is_parent_interface = False

    def __get_class_attrs(self):
        method_list = [method for method in dir(
            self._class) if method not in dir(self)]
        return method_list

    def __get_class_attr_types(self):
        attrs = self.__get_class_attrs()
        method_list = [func for func in attrs if callable(
            getattr(self._class, func))]
        variables = [func for func in attrs if not callable(
            getattr(self._class, func))]
        return method_list, variables

    def __is_ok(self):
        if not self._parent:
            return True
        attrs = self.__get_class_attr_types()
        parent_attrs = self._parent.__get_class_attr_types()
        for methods in attrs[0]:
            this_sig = signature(getattr(self._class, methods))
            this_params = this_sig.parameters
            parent_sig = signature(getattr(self._parent._class, methods))
            parent_params = parent_sig.parameters
            if (this_params) == parent_params:
                continue
            return f"Parameters Of Methods Are Not Equal (child( {' '.join(this_params)} ) , parent( {' '.join(parent_params) }))"
        for variable in parent_attrs[1]:
            if self.has_attr(variable):
                continue
            return f"Variable Problem. ( {variable} Not Defined In Child Class )"
        return True

    def has_attr(self, attr):
        """
            Checks If This Class Has Specific Attribute Or Not. ( Checks For Every Attribute Like Methods,Variables,... )
        """
        if attr in self.__get_class_attrs():
            return True
        return False

    def is_allowed_structure(self):
        """
            Checks If Has Allowed Structure Or Not.
        """
        is_ok = self.__is_ok()
        if is_ok != True:
            return is_ok
        return True

    def __call__(self, *parent_args, **parent_kwds):
        if self.is_parent_interface:
            return self._class(*parent_args, **parent_kwds)
        else:
            def wrapper(*args, **kwds):
                self._class = parent_args[0]
                interface.instances[self._class.__name__] = self._class
                try:
                    is_allowed = self.is_allowed_structure(self._parent)
                except:
                    raise ValueError("There Is No Interface With This Name.")
                if is_allowed != True:
                    raise ValueError(
                        f"This Class Does Not Extends From {self._class.__name__} Interface. {is_allowed}")
                return self._class(*args, **kwds)
            return wrapper


class Cond:
    """
Thats Just Normal One line Conditions That Exists in C++. ( Inspired From C++ )
You Can Use It Like : 
* Cond( condition )( if_its_true, if_its_false )
For Example :
* Cond(i > j)(i,j)
Or Cool Things Like :
* Cond(i > j)(minus,add)(i,j)
Thats Your Creativity That Makes Your Application Cool :)
    """

    def __init__(self, condition) -> None:
        self.condition = condition

    def __call__(self, t1, t2):
        if self.condition:
            return t1
        else:
            return t2


class Loop:
    """
This Option Is Unique Too Pysha. Its Not From Anywhere Else But Anyway I Think Its Not That Cool :/
You Can Use It Like : 
```Loop(x,y)("Enter Number[_1_][_2_] :",Loop.Modes.Input) # Input Mode Is Default```

### Modes : 
* Input = 0
* Print = 1

Input Is Where You Want To Get Input Just Not Print The Text In x * y Times. It Will Return x * y Array Contains Inputs.
Print Is Where You Want To Just Print The Text In x * y Times.
    """

    def __init__(self, *args):
        try:
            self.x = list(map(lambda x: int(x) - 1, args))
        except:
            raise ValueError("Values Must Be Int.")
        self.var_names = self.__get_var_name()
        self.index = 0
        self.ans_array = [None for _ in range(self.x[-1])]
        for i in self.x[:-1][::-1]:
            self.ans_array = [self.ans_array for _ in range(i)]

    def __get_var_name(self):
        self.index += 1
        return str(self.index)

    class Modes(Enum):
        Input = 0
        Print = 1

    def __format_text(self, text, from_replace=[], to_replace=[]):
        text = PyshaString(text)
        for c, i in enumerate(from_replace):
            text = text.replace(i, str(to_replace[c]))
        return text

    def __call__(self, text, mode=Modes.Input):
        cur = len(self.x) - 1
        self.y = [0 for _ in range(cur+1)]
        while True:
            text = self.__format_text(text, self.var_names, self.y)
            if mode == self.Modes.Input:
                set_array(self.ans_array, self.y, input(text))
            else:
                print(text)
            if self.y[cur] < self.x[cur]:
                self.y[cur] += 1
            elif list(self.x) == list(self.y):
                break
            else:
                self.y[cur] = 0
                ind = cur - 1
                while self.y[ind] == self.x[cur]:
                    self.y[ind] = 0
                    ind -= 1
                    if ind == 0:
                        break
                self.y[ind] += 1
        if mode == self.Modes.Input:
            return self.ans_array



class LoopError(Exception):
    pass


class ModeError(Exception):
    pass


class ConditionError(Exception):
    pass


class StackError(Exception):
    pass


class QueueError(Exception):
    pass


class SecurityError(Exception):
    pass


class AlgorithmError(Exception):
    pass


def exception_create(name):
    """
        Create Your Customized Exception With Name That You Specify !
    """
    return type(name, (Exception,), {})



from enum import *


class Algorithms(Enum):
    """
        The Algorythms That You Can Make Encryption Chain From.
    """
    XOR = 0
    Base64 = 1
    Cipher = 2
    Base85 = 3
    Base32 = 4
    Base16 = 5
    ROT13 = 6



from colorama import Fore, Back

fore = {
    "green": Fore.GREEN,
    "black": Fore.BLACK,
    "red": Fore.RED,
    "yellow": Fore.YELLOW,
    "blue": Fore.BLUE,
    "cyan": Fore.CYAN,
    "light_black": Fore.LIGHTBLACK_EX,
    "light_blue": Fore.LIGHTBLUE_EX,
    "light_cyan": Fore.LIGHTCYAN_EX,
    "light_green": Fore.LIGHTGREEN_EX,
    "light_magenta": Fore.LIGHTMAGENTA_EX,
    "light_red": Fore.LIGHTRED_EX,
    "light_white": Fore.LIGHTWHITE_EX,
    "light_yellow": Fore.LIGHTYELLOW_EX,
    "magenta": Fore.MAGENTA,
    "reset": Fore.RESET,
    "white": Fore.WHITE
}
"""This Is fore Dictionary That Contains Lots Of Colors That You Can Use For Text Color!"""

back = {
    "green": Back.GREEN,
    "black": Back.BLACK,
    "red": Back.RED,
    "yellow": Back.YELLOW,
    "blue": Back.BLUE,
    "cyan": Back.CYAN,
    "light_black": Back.LIGHTBLACK_EX,
    "light_blue": Back.LIGHTBLUE_EX,
    "light_cyan": Back.LIGHTCYAN_EX,
    "light_green": Back.LIGHTGREEN_EX,
    "light_magenta": Back.LIGHTMAGENTA_EX,
    "light_red": Back.LIGHTRED_EX,
    "light_white": Back.LIGHTWHITE_EX,
    "light_yellow": Back.LIGHTYELLOW_EX,
    "magenta": Back.MAGENTA,
    "reset": Back.RESET,
    "white": Back.WHITE
}
"""This Is back Dictionary That Contains Lots Of Colors That You Can Use For Background Of Text's Color!"""
